/*
 * IRC - Internet Relay Chat, ircd/s_bdd.c
 * Copyright (C) 1999 IRC-Hispano.org - ESNET - jcea & savage
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "sys.h"
#include <stdlib.h>

#if defined(BDD)
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <syslog.h>
#include "h.h"
#include "struct.h"
#include "ircd.h"
#include "s_serv.h"
#include "s_misc.h"
#include "sprintf_irc.h"
#include "send.h"
#include "s_err.h"
#include "numeric.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "hash.h"
#include "common.h"
#include "match.h"
#include "crule.h"
#include "parse.h"
#include "numnicks.h"
#include "userload.h"
#include "s_user.h"
#include "channel.h"
#include "querycmds.h"
#include "IPcheck.h"

#include "s_bdd.h"

#if defined(ESNET_NEG) && defined(ZLIB_ESNET)
#include "dbuf.h"
#endif

#if defined(BDD)
#include "msg.h"
#include "support.h"

/*
 * Las tablas con los registros, serie, version ...
 */

/* DB_HASH_SIZE ha de ser 2^n */
#define DB_HASH_SIZE      256
#define DB_MAX_TABLA      256

/*
** Numero de peticiones de registros
** antes de que haya que empezar a reciclar
*/
#define DB_BUF_CACHE  32

static struct db_reg db_buf_cache[DB_BUF_CACHE];
static int db_buf_cache_i = 0;

struct tabla_en_memoria {
  char *posicion;
  unsigned int len;
  char *puntero_r;
  char *puntero_w;
};

static int tabla_residente[DB_MAX_TABLA];
static unsigned int tabla_cuantos[DB_MAX_TABLA];
static struct db_reg *tabla_datos[DB_MAX_TABLA][DB_HASH_SIZE];
static unsigned int tabla_version[DB_MAX_TABLA];
static unsigned int tabla_serie[DB_MAX_TABLA];
static unsigned int tabla_hash_hi[DB_MAX_TABLA];
static unsigned int tabla_hash_lo[DB_MAX_TABLA];

int db_es_residente(char tabla)
{
  return tabla_residente[tabla];
}

unsigned int db_num_serie(char tabla)
{
  return tabla_serie[tabla];
}

unsigned int db_cuantos(char tabla)
{
  return tabla_cuantos[tabla];
}

#endif /* BDD */


#if defined(BDD)
/*
 * TEA (cifrado)
 *
 * Cifra 64 bits de datos, usando clave de 64 bits (los 64 bits superiores son cero)
 * Se cifra v[0]^x[0], v[1]^x[1], para poder hacer CBC facilmente.
 *
 */
void tea(unsigned int v[], unsigned int k[], unsigned int x[])
{
  unsigned int y = v[0] ^ x[0], z = v[1] ^ x[1], sum = 0, delta = 0x9E3779B9;
  unsigned int a = k[0], b = k[1], n = 32;
  unsigned int c = 0, d = 0;

  while (n-- > 0)
  {
    sum += delta;
    y += (z << 4) + a ^ z + sum ^ (z >> 5) + b;
    z += (y << 4) + c ^ y + sum ^ (y >> 5) + d;
  }

  x[0] = y;
  x[1] = z;
}

static void actualiza_hash(char *registro, char que_bdd)
{
  unsigned int buffer[129 * sizeof(unsigned int)];
  unsigned int *p = buffer;
  unsigned int x[2], v[2], k[2];
  char *p2;

/*
** Calculamos el HASH
*/
  memset(buffer, 0, sizeof(buffer));
  strncpy((char *)buffer, registro, sizeof(buffer) - 1);
  while ((p2 = strchr((char *)buffer, '\n')))
    *p2 = '\0';
  while ((p2 = strchr((char *)buffer, '\r')))
    *p2 = '\0';
  k[0] = k[1] = 0;
  x[0] = tabla_hash_hi[que_bdd];
  x[1] = tabla_hash_lo[que_bdd];
  while (*p)
  {
    v[0] = ntohl(*p++);
    v[1] = ntohl(*p++);
    tea(v, k, x);
  }
  tabla_hash_hi[que_bdd] = x[0];
  tabla_hash_lo[que_bdd] = x[1];
}

static void segmentar(char *registro, char **num_serie, char **destino,
    char **bdd, char **clave, char **clave_fin, char **valor, char **valor_fin)
{
  char *p;

  if (destino)
    *destino = NULL;
  if (bdd)
    *bdd = NULL;
  if (clave)
    *clave = NULL;
  if (clave_fin)
    *clave_fin = NULL;
  if (valor)
    *valor = NULL;
  if (valor_fin)
    *valor_fin = NULL;

  if (num_serie)
    *num_serie = p = registro;
  p = strchr(p, ' ');
  if (!p)
    return;
  if (destino)
    *destino = ++p;
  p = strchr(p, ' ');
  if (!p)
    return;
  if (bdd)
    *bdd = ++p;
  p = strchr(p, ' ');
  if (!p)
    return;
  if (clave)
    *clave = ++p;
  while ((*p != '0') && (*p != ' ') && (*p != '\n'))
    p++;
  if (clave_fin)
    *clave_fin = p;
  if (*p != ' ')
    return;
  if (valor)
    *valor = ++p;
  while ((*p != '\0') && (*p != '\n'))
    p++;
  if (valor_fin)
    *valor_fin = p;
}

/*
** db_busca_db_reg
*/
static struct db_reg *db_busca_db_reg(char tabla, char *clave)
{
  static char *c = NULL;
  static int c_len = 0;
  int i, hashi;
  char *p, *p2;
  struct db_reg *reg;

  if ((strlen(clave) + 1 > c_len) || (!c))
  {
    c_len = strlen(clave) + 1;
    if (c)
      RunFree(c);
    c = RunMalloc(c_len);
    if (!c)
      return;
  }
  strcpy(c, clave);
  /* paso a minusculas */
  i = 0;
  while (c[i] != 0)
  {
    c[i] = toLower(c[i]);
    i++;
  }
  hashi = db_hash_registro(c, DB_HASH_SIZE);

  for (reg = tabla_datos[tabla][hashi]; reg != NULL; reg = reg->next)
  {
/*
      segmentar(reg->p,NULL,NULL,NULL,&p,&p2,NULL,NULL);
      if ((strlen(c)==p2-p) && (!strncmp (p, c,p2-p)))
        return reg;
*/
    if (!strcmp(reg->clave, c))
      return reg;
  }
  return NULL;
}

/*
 * db_eliminar_registro (tabla, clave)
 *
 * libera (free) un registro de memoria.
 *                                      1999/06/23 savage@apostols.org
 */
static void db_eliminar_registro(char tabla, char *clave)
{
  struct db_reg *reg, *reg2, **reg3;
  int hashi, i = 0;
  static char *c = NULL;
  static int c_len = 0;

  if ((strlen(clave) + 1 > c_len) || (!c))
  {
    c_len = strlen(clave) + 1;
    if (c)
      RunFree(c);
    c = RunMalloc(c_len);
    if (!c)
      return;
  }
  strcpy(c, clave);
  /* paso a minusculas */
  while (c[i] != 0)
  {
    c[i] = toLower(c[i]);
    i++;
  }

  hashi = db_hash_registro(c, DB_HASH_SIZE);

  reg3 = &tabla_datos[tabla][hashi];

  for (reg = *reg3; reg != NULL; reg = reg2)
  {
    reg2 = reg->next;
    if (!strcmp(reg->clave, c))
    {
      *reg3 = reg2;

      if (reg->clave)
	RunFree(reg->clave);
      if (reg->valor)
	RunFree(reg->valor);
      RunFree(reg);
      tabla_cuantos[tabla]--;
      break;
    }
    reg3 = &(reg->next);
  }
}

/*
 * db_insertar_registro (tabla, clave, valor)
 *
 * mete un registro en memoria ...
 *                                      1999/06/23 savage@apostols.org
 */
static void db_insertar_registro(char tabla, char *clave, char *valor)
{
  struct db_reg *reg;
  int hashi;
  char *c, *v;
  aClient *acptr;
  int i = 0;

  /* lo borro primero, por si es un cambio */
  db_eliminar_registro(tabla, clave);

  c = RunMalloc(strlen(clave) + 1);
  if (!c)
    return;
  v = RunMalloc(strlen(valor) + 1);
  if (!v)
  {
    RunFree(c);
    return;
  }

  reg = RunMalloc(sizeof(struct db_reg));
  if (!reg)
  {
    RunFree(c);
    RunFree(v);
    return;
  }

  strcpy(c, clave);
  strcpy(v, valor);

  /* paso a minusculas */
  while (c[i] != 0)
  {
    c[i] = toLower(c[i]);
    i++;
  }

  /* creo el registro */
  reg->clave = c;
  reg->valor = v;
  reg->next = NULL;

  /* busco hash */
  hashi = db_hash_registro(reg->clave, DB_HASH_SIZE);

  /*
     sendto_ops("Inserto T='%c' C='%s' H=%u",tabla, reg->clave, hashi );
   */

  reg->next = tabla_datos[tabla][hashi];
  tabla_datos[tabla][hashi] = reg;

  tabla_cuantos[tabla]++;
}

static void copia_en_malloc(char *buf, int len, char **p)
{
  char *p2;

  p2 = *p;
  if ((p2 != NULL) && (strlen(p2) < len))
  {
    RunFree(p2);
    p2 = NULL;
  }
  if (!p2)
  {
    p2 = RunMalloc(len + 1);	/* El '\0' final */
    *p = p2;
  }
  memcpy(p2, buf, len);
  p2[len] = '\0';
}


struct db_reg *db_buscar_registro(char tabla, char *clave)
{
  struct db_reg *reg;
  char *clave_init, *clave_end;
  char *valor_init, *valor_end;
  char **p, *p2;

  reg = db_busca_db_reg(tabla, clave);
  if (!reg)
    return NULL;

/*
  segmentar(reg->p,NULL,NULL,NULL,&clave_init,&clave_end,&valor_init,&valor_end);
*/
/* Lo que sigue lo sustituye */
  clave_init = reg->clave;
  clave_end = clave_init + strlen(clave_init);
  valor_init = reg->valor;
  valor_end = valor_init + strlen(valor_init);

  copia_en_malloc(clave_init, clave_end - clave_init,
      &db_buf_cache[db_buf_cache_i].clave);
  copia_en_malloc(valor_init, valor_end - valor_init,
      &db_buf_cache[db_buf_cache_i].valor);
  if (++db_buf_cache_i >= DB_BUF_CACHE)
    db_buf_cache_i = 0;

  return reg;
}

/*
 * db_es_miembro (tabla, clave, subcadena)
 *
 * varor registro es una lista separada por comas, y si subcadena es
 # una de ellas, retorna la posicion, sino 0
 *                                      1999/07/03 savage@apostols.org
 */
int db_es_miembro(char tabla, char *clave, char *subcadena)
{
  int j, i = 0;
  static char *buf = NULL;
  static int buf_len = 0;
  char *f, *s = NULL;
  struct db_reg *reg;

  if ((reg = db_buscar_registro(tabla, clave)) == NULL)
    return 0;

  if ((strlen(reg->valor) + 1 > buf_len) || (!buf))
  {
    buf_len = strlen(reg->valor) + 1;
    if (buf)
      RunFree(buf);
    buf = RunMalloc(buf_len);
    if (!buf)
      return 0;
  }

  strcpy(buf, reg->valor);
  for (f = strtoken(&s, buf, ","); f != NULL; f = strtoken(&s, NULL, ","))
  {
    j++;
    if (!strCasediff(f, subcadena))
    {
      i++;
      break;
    }
  }

  return i;
}

#endif /* defined(BDD) */

/*
** Esta rutina mata el servidor.
** Es casi una copia de m_die().
*/
extern char **myargv;		/* ircd.c */
void db_die(char *msg, char que_bdd)
{
  aClient *acptr;
  int i;
  char buf[1024];

  sprintf_irc(buf, "DB '%c' - %s (%s). El daemon muere...", que_bdd, msg,
      DBPATH);
  for (i = 0; i <= highest_fd; i++)
  {
    if (!(acptr = loc_clients[i]))
      continue;
    if (IsUser(acptr))
      sendto_one(acptr, ":%s NOTICE %s :%s", me.name, acptr->name, buf);
    else if (IsServer(acptr))
      sendto_one(acptr, ":%s ERROR :%s", me.name, buf);
  }

#ifndef  USE_SYSLOG
  openlog(myargv[0], LOG_PID | LOG_NDELAY, LOG_DAEMON);
#endif

  syslog(LOG_ERR, buf);

#ifndef  USE_SYSLOG
  closelog();
#endif

  Debug((DEBUG_ERROR, buf));

#ifdef __cplusplus
  s_die(0);
#else
  s_die();
#endif
}

#ifdef BDD
/*
 * leer_db
 *
 * Lee un registro de la base de datos.
 *
 */
static int leer_db(struct tabla_en_memoria *mapeo, char *buf)
{
  int cont = 0;
  int estado;
  char *p = mapeo->posicion + mapeo->len;

  while (mapeo->puntero_r < p)
  {
    *buf = *(mapeo->puntero_r++);
    if (*buf == '\r')
      continue;
    if ((*buf++ == '\n'))
    {
      *--buf = '\0';
      return cont;
    }
    if (cont++ > 500)
      break;
  }
  *buf = '\0';
  return -1;
}

/*
** Seek por bidivision
*/
static int
seek_db(struct tabla_en_memoria *mapeo, char *buf, unsigned int registro)
{
  char *p, *p2, *plo, *phi;
  unsigned int v;

/*
** Este caso especial es lo bastante 
** comun como para que se tenga en cuenta.
*/
  if (!registro)
  {
    mapeo->puntero_r = mapeo->posicion;
    return leer_db(mapeo, buf);
  }

  plo = mapeo->posicion;
  phi = plo + mapeo->len;
  while (plo != phi)
  {
/*
** p apunta al principio del registro
** p2 apunta al registro siguiente
*/
    p2 = p = plo + (phi - plo) / 2;
    while ((p >= plo) && (*p != '\n'))
      p--;
    if (p < plo)
      p = plo;
    if (*p == '\n')
      p++;
    while ((p2 < phi) && (*p2++ != '\n'));
/*
** Estamos al principio de un registro
*/
    v = atol(p);
    if (v > registro)
      phi = p;
    else if (v < registro)
      plo = p2;
    else
    {
      plo = phi = p;		/* Encontrado */
      break;
    }
  }
  mapeo->puntero_r = plo;
  return leer_db(mapeo, buf);
}

/*
 * abrir_db
 *
 * Se mueve hasta un registro de la base de datos
 *
 */
static int
abrir_db(unsigned int registro, char *buf, char que_bdd,
    struct tabla_en_memoria *mapeo)
{
  int handle;
  int cont;
  char path[1024];
  struct stat estado;

  *buf = '\0';
  sprintf_irc(path, "%s/tabla.%c", DBPATH, que_bdd);
  alarm(3);
  handle = open(path, O_RDONLY, S_IRUSR | S_IWUSR);
  fstat(handle, &estado);
  mapeo->len = estado.st_size;
  mapeo->posicion = mmap(NULL, mapeo->len,
      PROT_READ, MAP_SHARED | MAP_NORESERVE, handle, 0);
  close(handle);
  alarm(0);
  mapeo->puntero_r = mapeo->puntero_w = mapeo->posicion;

  if (handle == -1)
    db_die("Error al intentar leer (open)", que_bdd);
  if ((mapeo->len != 0) && (mapeo->posicion == MAP_FAILED))
    db_die("Error al intentar leer (mmap)", que_bdd);

  if ((handle == -1) || (mapeo->posicion == MAP_FAILED))
    return -1;

  cont = seek_db(mapeo, buf, registro);
  if (cont == -1)
  {
    *buf = '\0';
    return -1;
  }
  return handle;
}

static void cerrar_db(struct tabla_en_memoria *mapeo)
{
  munmap(mapeo->posicion, mapeo->len);
}

/*
** almacena_hash
**
** El fichero de HASHES contiene una linea
** por BDD. Cada linea tiene el formato
** "BDD HASH\n"
** En total, cada linea mide 15 bytes.
*/
static void almacena_hash(char que_bdd)
{
  char path[1024];
  char hash[20];
  int db_file;

  sprintf_irc(path, "%s/hashes", DBPATH);
  inttobase64(hash, tabla_hash_hi[que_bdd], 6);
  inttobase64(hash + 6, tabla_hash_lo[que_bdd], 6);
  alarm(3);
  db_file = open(path, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
  if (db_file == -1)
    db_die("Error al intentar guardar hashes (open)", que_bdd);
  sprintf_irc(path, "%c %s\n", que_bdd, hash);
  if (lseek(db_file, 15 * (que_bdd - ESNET_BDD), SEEK_SET) == -1)
    db_die("Error al intentar guardar hashes (lseek)", que_bdd);
  if (write(db_file, path, strlen(path)) == -1)
    db_die("Error al intentar guardas hashes (write)", que_bdd);
  close(db_file);
  alarm(0);
}

/*
** lee_hash
*/
static void lee_hash(char que_bdd, unsigned int *hi, unsigned int *lo)
{
  char path[1024];
  char c;
  int db_file;

  sprintf_irc(path, "%s/hashes", DBPATH);
  alarm(3);
  db_file = open(path, O_RDONLY);
/*
** No metemos verificacion, porque ya verifica
** al contrastar el hash.
*/
  lseek(db_file, 15 * (que_bdd - ESNET_BDD) + 2, SEEK_SET);
  read(db_file, path, 12);
  close(db_file);
  alarm(0);
  path[12] = '\0';
  c = path[6];
  path[6] = '\0';
  *hi = base64toint(path);
  path[6] = c;
  *lo = base64toint(path + 6);
}

/*
 * db_alta
 *
 * Da de alta una entrada en la base de datos en memoria
 * Formato "serie destino id clave contenido"
 *
 * modificado para usar las hash con funciones db_*
 *                                      1999/06/30 savage@apostols.org
 */
static void db_alta(char *registro, char que_bdd, int guarda_disco)
{
  char *p0, *p1, *p2, *p3, *p4;
  char path[1024];
  int db_file;

  actualiza_hash(registro, que_bdd);

  if (guarda_disco)
  {
    almacena_hash(que_bdd);

    sprintf_irc(path, "%s/tabla.%c", DBPATH, que_bdd);
    alarm(3);
    db_file = open(path, O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR);
    if (db_file == -1)
      db_die("Error al intentar an~adir nuevo registro (open)", que_bdd);
    if (write(db_file, registro, strlen(registro)) == -1)
      db_die("Error al intentar an~adir nuevo registro (write)", que_bdd);
    close(db_file);
    alarm(0);
  }

  p0 = strtok(registro, " ");	/* serie */
  p1 = strtok(NULL, " ");	/* destino */
  p2 = strtok(NULL, " ");	/* tabla */
  p3 = strtok(NULL, " \r\n");	/* clave */
  p4 = strtok(NULL, "\r\n");	/* valor (opcional) */

  if (p3 == NULL)
    return;			/* registro incompleto */

  tabla_serie[que_bdd] = atol(p0);

  if (tabla_residente[que_bdd])
  {
    if (p4 == NULL)		/* Borrado */
      db_eliminar_registro(que_bdd, p3);
    else
      db_insertar_registro(que_bdd, p3, p4);
  }
}

/*
** db_pack
**
** Elimina los registro superfluos
** de una Base de Datos Local.
** 
** Se invoca cuando se recibe un CheckPoint, y el
** formato es "serie destino id * texto"
*/
static void db_pack(char *registro, char que_bdd)
{
  int db_file;
  char path[1024];
  char *map;
  char *lectura, *escritura, *p;
  char *clave, *valor;
  char c;
  unsigned int len, len2;
  struct stat estado;
  struct db_reg *reg;

/*
** El primer valor es el numero de serie actual
*/
  tabla_serie[que_bdd] = atol(registro);

  sprintf_irc(path, "%s/tabla.%c", DBPATH, que_bdd);
  alarm(3);
  db_file = open(path, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
  fstat(db_file, &estado);
  len = estado.st_size;
  map = mmap(NULL, len, PROT_READ | PROT_WRITE,
      MAP_SHARED | MAP_NORESERVE, db_file, 0);
  close(db_file);
  alarm(0);
  if (db_file == -1)
    db_die("Error al intentar compactar (open)", que_bdd);
  if ((len != 0) && (map == MAP_FAILED))
    db_die("Error al intentar compactar (mmap)", que_bdd);

  if (!tabla_residente[que_bdd])
  {				/* No residente -> No pack */
    escritura = map + len;
    goto fin;
  }

  tabla_hash_hi[que_bdd] = 0;
  tabla_hash_lo[que_bdd] = 0;

  p = lectura = escritura = map;
  while (lectura < map + len)
  {
    p = strchr(p, ' ') + 1;	/* Destino */
    p = strchr(p, ' ') + 1;	/* BDD */
    valor = clave = p = strchr(p, ' ') + 1;	/* Clave */
    while ((p < map + len) && (*p != '\n') && (*p != '\r'))
      p++;
    while ((*valor != ' ') && (valor < p))
      valor++;
    valor++;			/* Nos saltamos el espacio */

/*
** Los registros "*" son borrados automaticamente
** al compactar, porque no aparecen en la
** base de datos en memoria.
**
** Solo se mantiene el nuevo, porque es
** el que se graba tras la compactacion
*/
    if (valor < p)
    {				/* Nuevo registro, no un borrado */
      *(valor - 1) = '\0';
      reg = db_buscar_registro(que_bdd, clave);
      *(valor - 1) = ' ';
      if (reg != NULL)
      {				/* El registro sigue existiendo */

	len2 = strlen(reg->valor);
	if ((valor + len2 == p) && (!strncmp(valor, reg->valor, len2)))
	{			/* !!El mismo!! */

/*
** Actualizamos HASH
*/
	  if ((*p == '\n') || (*p == '\r'))
	  {
	    c = *p;
	    *p = '\0';
	    actualiza_hash(lectura, que_bdd);
	    *p = c;
	  }
	  else
	  {			/* Estamos al final y no hay retorno de carro */
	    memcpy(path, lectura, p - lectura);
	    path[p - lectura] = '\0';
	    actualiza_hash(path, que_bdd);
	  }

	  while ((p < map + len) && ((*p == '\n') || (*p == '\r')))
	    p++;
	  memcpy(escritura, lectura, p - lectura);
	  escritura += p - lectura;
	  lectura = p;
	  continue;		/* MUY IMPORTANTE */
	}			/* Hay otro mas moderno que este */
      }				/* El registro fue borrado */
    }				/* Es un borrado */
    while ((p < map + len) && ((*p == '\n') || (*p == '\r')))
      p++;
    lectura = p;
  }

fin:

  munmap(map, len);

  alarm(3);
  if (truncate(path, escritura - map) == -1)
  {
    db_die("Error al intentar compactar (truncate)", que_bdd);
  }
  db_file = open(path, O_RDWR | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR);
  if (db_file == -1)
  {
    db_die("Error al intentar compactar (open-2)", que_bdd);
  }
  write(db_file, registro, strlen(registro));	/* CheckPoint */
  close(db_file);
  alarm(0);
  actualiza_hash(registro, que_bdd);
  almacena_hash(que_bdd);
}

/*
 * borrar_db
 *
 * Borra la base de datos en memoria (modificado para uso de hash)
 *
 *                                      1999/06/30 savage@apostols.org
 */
static void borrar_db(char que_bdd)
{
  int i;
  struct db_reg *reg, *reg2;

  for (i = 0; i < DB_HASH_SIZE; i++)
  {
    for (reg = tabla_datos[que_bdd][i]; reg != NULL; reg = reg2)
    {
      reg2 = reg->next;
      if (reg->clave)
	RunFree(reg->clave);
      if (reg->valor)
	RunFree(reg->valor);
      RunFree(reg);
    }
    tabla_datos[que_bdd][i] = NULL;
  }

/* Error introducido en DBH
** BUGFIX!
*/
  tabla_serie[que_bdd] = 0;
  tabla_cuantos[que_bdd] = 0;
  tabla_hash_hi[que_bdd] = 0;
  tabla_hash_lo[que_bdd] = 0;
}


static void corta_si_multiples_hubs(aClient *cptr, char que_bdd, char *mensaje)
{
  char buf[1024];
  Dlink *lp;
  int num_hubs = 0;
  aClient *acptr;

  for (lp = me.serv->down; lp; lp = lp->next)
  {
    if (find_conf_host(lp->value.cptr->confs,
	lp->value.cptr->name, CONF_HUB) != NULL)
      num_hubs++;
  }

  if (num_hubs >= 2)
  {

/*
** No podemos simplemente hace el bucle, porque
** el "exit_client()" modifica la estructura
*/
  corta:
    if (num_hubs-- > 1)
    {				/* Deja un HUB conectado */
      for (lp = me.serv->down; lp; lp = lp->next)
      {
	acptr = lp->value.cptr;
/*
** Si se especifica que se desea mantener un HUB
** en concreto, respeta esa peticion
*/
	if ((acptr != cptr) &&
	    find_conf_host(acptr->confs, acptr->name, CONF_HUB) != NULL)
	{
	  sprintf_irc(buf, "BDD '%c' %s. Resincronizando...", que_bdd, mensaje);
	  exit_client(acptr, acptr, &me, buf);
	  goto corta;
	}
      }
    }
  }
}

/*
 * initdb2
 *
 * Lee la base de datos de disco
 *
 */
static void initdb2(char que_bdd)
{
  unsigned int hi, lo;
  char buf[1024];
  char path[1024];
  int db_file;
  Dlink *lp;
  struct tabla_en_memoria mapeo;
  char *destino, *p = NULL, *p2, *p3;
  int p_len = 0;

  borrar_db(que_bdd);

  db_file = abrir_db(0, buf, que_bdd, &mapeo);
  if (db_file != -1)
    do
    {
      /* db_alta modifica la cadena */
      destino = NULL;
      if (tabla_residente[que_bdd])
      {
	destino = strchr(buf, ' ') + 1;
	p3 = strchr(destino, ' ') + 1;
	p2 = strchr(p3, ' ') + 1;
/*
** Curemonos en salud
*/
	if ((*(p3 + 1) != ' ') || ((*p3 != que_bdd) && !((*p3 == 'N')
	    && (que_bdd == 'n'))))
	{
/*
** HACK para que la verificacion HASH falle
** y asi purgar la base de datos defectuosa.
** Los valores usados son arbitrarios.
*/
	  tabla_hash_hi[que_bdd] += 57;
	  tabla_hash_lo[que_bdd] += 97;
	}
	if ((*destino == '*') && (*(destino + 1) == ' '))
	{
/*
** Este caso es tan normal
** que lo optimizamos.
*/
	  destino = NULL;
	}
	else
	{
	  if (strlen(destino) + 1 > p_len)
	  {
	    p_len = strlen(destino) + 1;
	    if (p)
	      RunFree(p);
	    p = RunMalloc(p_len);
	  }
	  strcpy(p, destino);
	  *strchr(p, ' ') = '\0';
	  collapse(p);
	  if (!match(p, me.name))
	    destino = NULL;	/* Para nosotros */
	}
      }
      if (tabla_residente[que_bdd] && (destino == NULL) && !((*p2 == '*')
	  && (*(p2 + 1) == '\0')))
      {
	db_alta(buf, que_bdd, 0);
      }
      else
      {
/*
** Necesitamos coger su numero de serie
** para poder determinar si los
** registros que nos lleguen por la red
** hay que guardarlos en disco, ademas
** de poder negociar convenientemente
** las versiones de las BDD
** en un NetJoin.
*/
	tabla_serie[que_bdd] = atol(buf);
	actualiza_hash(buf, que_bdd);
      }
    }
    while (leer_db(&mapeo, buf) != -1);
  cerrar_db(&mapeo);

  if (p)
    RunFree(p);

/*
** Ahora comprueba que el HASH de la BDD
** cargada se corresponda con el HASH almacenado
*/
  lee_hash(que_bdd, &hi, &lo);
  if ((tabla_hash_hi[que_bdd] != hi) || (tabla_hash_lo[que_bdd] != lo))
  {
    sendto_ops("ATENCION - Base de Datos "
	"'%c' aparentemente corrupta. Borrando...", que_bdd);
    borrar_db(que_bdd);
    sprintf_irc(path, "%s/tabla.%c", DBPATH, que_bdd);
    alarm(3);
    db_file = open(path, O_TRUNC, S_IRUSR | S_IWUSR);
    close(db_file);
    if (db_file == -1)
      db_die("Error al intentar truncar (open)", que_bdd);
    alarm(0);
/*
** Solucion temporal
** Corta conexiones con los HUBs
*/
  corta_hubs:
    for (lp = me.serv->down; lp; lp = lp->next)
    {
      if (find_conf_host(lp->value.cptr->confs,
	  lp->value.cptr->name, CONF_HUB) != NULL)
      {
	sprintf_irc(buf, "BDD '%c' inconsistente. Resincronizando...", que_bdd);
	exit_client(lp->value.cptr, lp->value.cptr, &me, buf);
#if 0
	/* fix patch.db52 */
	/* lp ya no es valido al desconectar */
	lp = me.serv->down;
	if (!lp)
	  break;
#endif
	/* fix patch.db74 */
	goto corta_hubs;
      }
    }
/*
** Fin solucion temporal
*/

    corta_si_multiples_hubs(NULL, que_bdd, "inconsistente");
    sendto_ops("Solicitando actualizacion BDD '%c' "
	"a los nodos vecinos.", que_bdd);
/*
** Solo pide a los HUBs, porque no se
** aceptan datos de leafs.
*/
    for (lp = me.serv->down; lp; lp = lp->next)
    {
      if (find_conf_host(lp->value.cptr->confs,
	  lp->value.cptr->name, CONF_HUB) != NULL)
      {
	sendto_one(lp->value.cptr, "%s DB %s 0 J %lu %c",
	    NumServ(&me), lp->value.cptr->name, tabla_serie[que_bdd], que_bdd);
      }
    }
  }
  almacena_hash(que_bdd);

/*
** Si hemos leido algun registro de compactado,
** sencillamente nos lo cargamos en memoria, para
** que cuando llegue otro, el antiguo se borre
** aunque tenga algun contenido (un texto explicativo, por ejemplo).
** Parche DB70
*/

  db_eliminar_registro(que_bdd, "*");
}

void initdb(void)
{
  char c;
  memset(tabla_residente, 0, sizeof(tabla_residente));
  tabla_residente[ESNET_NICKDB] = 1;
#if defined(BDD) && defined(BDD_CLONES)
  tabla_residente[ESNET_CLONESDB] = 1;
#endif
  tabla_residente[BDD_OPERDB] = 1;
  tabla_residente[BDD_CHANDB] = 1;
  tabla_residente[BDD_BOTSDB] = 1;
  tabla_residente[BDD_IPVIRTUALDB] = 1;
  tabla_residente[BDD_IPVIRTUAL2DB] = 1;
  for (c = ESNET_BDD; c <= ESNET_BDD_END; c++)
    initdb2(c);
}

/*
 * reload_db
 *
 * Recarga la base de datos de disco, liberando la memoria
 *
 */
void reload_db(void)
{
  char buf[16];
  char c;
  sendto_ops("Releyendo Bases de Datos...");
  initdb();
  for (c = ESNET_BDD; c <= ESNET_BDD_END; c++)
  {
    if (tabla_serie[c])
    {
      inttobase64(buf, tabla_hash_hi[c], 6);
      inttobase64(buf + 6, tabla_hash_lo[c], 6);
      sendto_ops("DB: '%c'. Ultimo registro: %lu. HASH: %s",
	  c, tabla_serie[c], buf);
    }
  }
}

void tx_num_serie_dbs(aClient *cptr)
{
  int cont;

/*
** Informamos del estado de nuestras BDD
*/
#if defined(ESNET_NEG) && defined(ZLIB_ESNET)
  inicia_microburst();
#endif

/* La tabla 'n' es un poco especial... */
  sendto_one(cptr, "%s DB * 0 J %lu 2", NumServ(&me),
      db_num_serie(ESNET_NICKDB));

  for (cont = ESNET_BDD; cont <= ESNET_BDD_END; cont++)
  {
    if (cont != ESNET_NICKDB)	/* No mandamos nicks de nuevo */
      sendto_one(cptr, "%s DB * 0 J %lu %c",
	  NumServ(&me), tabla_serie[cont], cont);
  }

#if defined(ESNET_NEG) && defined(ZLIB_ESNET)
  completa_microburst();
#endif
}


/*
 * m_db
 *
 * Gestion de la base de datos - ESNET
 * 29/May/98 jcea@argo.es
 *
 */
int m_db(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  unsigned int db;
  aClient *acptr;
  Dlink *lp;
  char db_buf[1024], db_buf2[1024];
  char path[1024];
  int db_file;
  int es_hub = 0;
  char *p, *p2, *p3, *p4;
  char que_bdd = ESNET_NICKDB;
  unsigned int mascara_bdd;
  int cont;
  struct tabla_en_memoria mapeo;
  if (!IsServer(sptr) || parc < 5)
    return 0;
  db = atol(parv[2]);
  if (find_conf_host(cptr->confs, cptr->name, CONF_HUB) != NULL)
    es_hub = !0;
  if (!db)
  {
    db = atol(parv[4]);
    if (parc == 6)
    {
      que_bdd = *parv[5];
      if ((que_bdd < 'a') || (que_bdd > 'z'))
      {
	if ((que_bdd == '*') && ((*parv[3] == 'Q') || (*parv[3] == 'R')))
	  que_bdd = '*';	/* Estamos preguntando por un HASH global de todas las tablas */
	else if ((que_bdd < '2') || (que_bdd > '9') || (*parv[3] != 'J'))
	  return 0;
	else
	  que_bdd = ESNET_NICKDB;
      }
    }

/*
** Hay CPUs (en particular, viejos INTEL) que
** si le dices 1372309 rotaciones, por ejemplo,
** pues te las hace ;-), con lo que ello supone de
** consumo de recursos. Lo normal es que
** la CPU haga la rotacion un numero de veces
** modulo 32 o 64, pero hay que curarse en salud.
*/
    if ((que_bdd >= ESNET_BDD) && (que_bdd <= ESNET_BDD_END))
      mascara_bdd = ((unsigned int)1) << (que_bdd - ESNET_BDD);

    switch (*parv[3])
    {
      case 'B':
	if (es_hub)
	  sendto_one(sptr, "%s DB %s 0 J %lu %c",
	      NumServ(&me), parv[0], tabla_serie[que_bdd], que_bdd);
	return 0;
	break;
      case 'J':
/*
** No hay problemas de CPU en el
** extremo receptor. Esto es para
** limitar el LAG y el caudal
** consumido.
*/
	cont = 1000;
	if (db >= tabla_serie[que_bdd])
	{			/* Se le pueden mandar registros individuales */
	  sptr->serv->esnet_db |= mascara_bdd;
	  return 0;
	  break;
	}
	else if ((sptr->serv->esnet_db) & mascara_bdd)
	{
/*
** Teniamos el grifo abierto, y ahora
** nos esta pidiendo registros antiguos.
** Eso SOLO ocurre si ha detectado que su
** copia local de la BDD esta corrupta.
*/
	  sptr->serv->esnet_db &= ~mascara_bdd;
/*
** Borramos su BDD porque es posible
** que le hayamos enviado mas registros.
** Es preferible empezar desde cero.
*/
	  sendto_one(sptr, "%s DB %s 0 D BDD_CORRUPTA %c",
	      NumServ(&me), sptr->name, que_bdd);
#if 0
	  sendto_one(sptr, "%s DB %s 0 B %lu %c",
	      NumServ(&me), sptr->name, tabla_serie[que_bdd], que_bdd);
#endif
	  /*
	   ** No enviamos nada porque ya nos lo pedira el
	   ** con un "J"
	   */
	  return;
	}

	db_file = abrir_db(db + 1, db_buf, que_bdd, &mapeo);
	if (db_file == -1)
	  return 0;		/* Problemas con la DB */
#if defined(ESNET_NEG) && defined(ZLIB_ESNET)
	inicia_microburst();
#endif
	do
	{
	  p = strchr(db_buf, ' ');
	  if (p == NULL)
	    continue;
	  *p++ = '\0';
	  p2 = strchr(p, ' ');
	  if (p2 == NULL)
	    continue;
	  *p2++ = '\0';
	  p3 = strchr(p2, ' ');
	  if (p3 == NULL)
	    continue;
	  *p3++ = '\0';
	  p4 = strchr(p3, ' ');
	  if (p4 == NULL)
	  {
	    sendto_one(sptr, "%s DB %s %s %s %s", NumServ(&me), p,
		db_buf, p2, p3);
	  }
	  else
	  {
	    *p4++ = '\0';
	    sendto_one(sptr, "%s DB %s %s %s %s :%s", NumServ(&me), p,
		db_buf, p2, p3, p4);
	  }
	  if (!(--cont))
	    break;
	}
	while (leer_db(&mapeo, db_buf) != -1);
#if defined(ESNET_NEG) && defined(ZLIB_ESNET)
	completa_microburst();
#endif
	cerrar_db(&mapeo);
	if (cont)
	  sptr->serv->esnet_db |= mascara_bdd;
	else
	  sendto_one(sptr, "%s DB %s 0 B %lu %c",
	      NumServ(&me), parv[0], tabla_serie[que_bdd], que_bdd);
	return 0;
	break;
      case 'D':
	if (!es_hub)
	  return 0;

/*
** Debemos enviar el broadcast ANTES de
** cortar los enlaces.
*/
	sprintf(db_buf, "%s DB %s 0 D %s %c",
	    NumServ(sptr), parv[1], parv[4], que_bdd);
	for (lp = me.serv->down; lp; lp = lp->next)
	{
	  if (lp->value.cptr == cptr)
	    continue;
/*
** No sabemos si la otra punta
** va a cumplir la mascara, asi que
** nos curamos en salud.
*/
	  lp->value.cptr->serv->esnet_db &= ~mascara_bdd;
	  sendto_one(lp->value.cptr, db_buf);
	}

/*
** Por bug en lastNNServer no se puede usar 'find_match_server()'
*/
	collapse(parv[1]);
	if (!match(parv[1], me.name))
	{
	  sprintf_irc(path, "%s/tabla.%c", DBPATH, que_bdd);
	  alarm(3);
	  db_file = open(path, O_TRUNC, S_IRUSR | S_IWUSR);
	  close(db_file);
	  if (db_file == -1)
	    db_die("Error al intentar truncar (open)", que_bdd);
	  alarm(0);
	  borrar_db(que_bdd);
	  almacena_hash(que_bdd);
	  sprintf_irc(db_buf, "borrada (%s)", sptr->name);
	  corta_si_multiples_hubs(cptr, que_bdd, db_buf);
/*
** Podemos enviar a 'sptr' tranquilos, porque el
** corte afectaria a todos los HUBs menos a
** traves del que llega la peticion.
*/
	  sendto_one(sptr, "%s DB %s 0 E %s %c\n",
	      NumServ(&me), sptr->name, parv[4], que_bdd);
	}

/*
** Como no sabemos si el otro extremo nos ha
** cerrado el grifo o no, nos curamos en salud
*/
	sendto_one(cptr, "%s DB %s 0 J %lu %c",
	    NumServ(&me), cptr->name, tabla_serie[que_bdd], que_bdd);

	return 0;
	break;
      case 'E':		/* Podemos ser destinatarios si el otro estaba corrupto */
      case 'R':
	if ((acptr = find_match_server(parv[1])) && (!IsMe(acptr)))
	  sendto_one(acptr, "%s DB %s 0 %c %s %c", NumServ(sptr),
	      acptr->name, *parv[3], parv[4], que_bdd);
	return 0;
	break;
      case 'Q':
	if (!es_hub)
	  return 0;
/*
** Por bug en lastNNServer no se puede usar 'find_match_server()'
*/
	collapse(parv[1]);
	if (!match(parv[1], me.name))
	{
	  if (que_bdd == '*')
	  {			/* Estamos preguntando por un HASH global de todas las tablas */
	    unsigned long int hashes_hi = 0;
	    unsigned long int hashes_lo = 0;
	    unsigned long tablas_series = 1;
	    unsigned long tablas_cuantos = 1;
	    int i;

	    for (i = ESNET_BDD; i <= ESNET_BDD_END; i++)
	    {
	      hashes_hi ^= tabla_hash_hi[i];
	      hashes_lo ^= tabla_hash_lo[i];
/*
** Lo que sigue da valores distintos si la
** maquina no es "complemento a dos", pero eso
** no es un problema en la red actual.
**
** El AND es para que salgan resultados
** consistentes tanto en maquinas 32 bits como 64 bits.
*/
	      tablas_series = (tablas_series * (tabla_serie[i] + 1)) & 0xfffffffful;	/* Por si el valor es cero */
	      tablas_cuantos = (tablas_cuantos * (tabla_cuantos[i] + 1)) & 0xfffffffful;	/* Por si el valor es cero */
	    }
	    inttobase64(db_buf, hashes_hi, 6);
	    inttobase64(db_buf + 6, hashes_lo, 6);
	    sendto_one(sptr, "%s DB %s 0 R %lu-%lu-%s-%s %c",
		NumServ(&me), sptr->name, tablas_series,
		tablas_cuantos, db_buf, parv[4], que_bdd);
	  }
	  else
	  {			/* Una tabla en particular */
	    inttobase64(db_buf, tabla_hash_hi[que_bdd], 6);
	    inttobase64(db_buf + 6, tabla_hash_lo[que_bdd], 6);
	    sendto_one(sptr, "%s DB %s 0 R %lu-%lu-%s-%s %c",
		NumServ(&me), sptr->name, tabla_serie[que_bdd],
		tabla_cuantos[que_bdd], db_buf, parv[4], que_bdd);
	  }
	}

	sprintf(db_buf, "%s DB %s 0 Q %s %c",
	    NumServ(sptr), parv[1], parv[4], que_bdd);
	for (lp = me.serv->down; lp; lp = lp->next)
	{
	  if (lp->value.cptr == cptr)
	    continue;
	  sendto_one(lp->value.cptr, db_buf);
	}

	return 0;
	break;
    }
    return 0;
  }

/* Nuevo registro */

  que_bdd = *parv[3];
/*
** Solo se admite un caracter
*/
  if (*(parv[3] + 1) != '\0')
    return 0;

  if ((que_bdd < 'a') || (que_bdd > 'z'))
  {
    if (que_bdd == 'N')
      que_bdd = ESNET_NICKDB;
    else
      return 0;
  }

  mascara_bdd = ((unsigned int)1) << (que_bdd - ESNET_BDD);
  if (db <= tabla_serie[que_bdd])
    return 0;
  if (!es_hub)
    return 0;
/*
** Ojo, hay que usar 'db' y no 'parv[2]'
** porque en la cadena de formateo se
** dice que se le pasa un numero, no
** una cadena.
** Se le pasa un numero para que se formatee
** correctamente, ya que luego sera
** utilizado en un HASH y todos los
** nodos deberian dar el mismo valor.
*/
  if (parc == 5)
    sprintf_irc(db_buf, "%s DB %s %lu %s %s",
	NumServ(sptr), parv[1], db, parv[3], parv[4]);
  else
    sprintf_irc(db_buf, "%s DB %s %lu %s %s :%s",
	NumServ(sptr), parv[1], db, parv[3], parv[4], parv[5]);
  for (lp = me.serv->down; lp; lp = lp->next)
  {
    if (!((lp->value.cptr->serv->esnet_db) & mascara_bdd) ||
	(lp->value.cptr == cptr))
      continue;
    sendto_one(lp->value.cptr, db_buf);
  }

  if (parc == 5)
    sprintf_irc(db_buf, "%lu %s %s %s\n", db, parv[1], parv[3], parv[4]);
  else
    sprintf_irc(db_buf, "%lu %s %s %s %s\n", db, parv[1], parv[3],
	parv[4], parv[5]);
  if (strcmp(parv[4], "*"))
  {
    db_alta(db_buf, que_bdd, !0);
  }
  else
  {				/* Checkpoint */
    db_pack(db_buf, que_bdd);
  }
  return 0;
}

#endif /* BDD */

#ifdef BDD

/*
 * m_dbq
 *
 *   parv[0]   = prefijo del enviador
 *   parv[1..] = argumentos de DBQ [<server>] <tabla> <clave>
 *
 * Maneja los mensajes DBQ recibidos de users y servers.
 *                                      1999/10/13 savage@apostols.org
 */
int m_dbq(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  char tabla, *clave, *servidor;
  aClient *acptr;
  struct db_reg *reg;
  int nivel_helper = 0;

  if (!IsUser(sptr))
    return 0;			/* Ignoramos peticiones de nodos */

  if (!IsOper(sptr) && !IsHelpOp(sptr))
  {
    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
    return 0;			/* No autorizado */
  }
  /* <Origen> DBQ [<server>] <Tabla> <Clave> */
  if (parc != 3 && parc != 4)
  {

    if (!IsServer(sptr))
      sendto_one(cptr,
	  ":%s NOTICE %s :Parametros incorrectos: Formato: DBQ [<server>] <Tabla> <Clave>",
	  me.name, parv[0]);
    return 0;
  }

  if (parc == 3)
  {
    servidor = NULL;		/* no nos indican server */
    tabla = *parv[1];
    clave = parv[2];
  }
  else
  {
    servidor = parv[1];
    tabla = *parv[2];
    clave = parv[3];
    if (*servidor == '*')
    {
      /* WOOW, BROADCAST */
      sendto_serv_butone(cptr, ":%s DBQ * %c %s", parv[0], tabla, clave);
    }
    else
    {
      /* NOT BROADCAST */
      if (!(acptr = find_match_server(servidor)))
      {
	/* joer, el server de destino no existe */
	sendto_one(cptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], servidor);
	return 0;
      }

      if (!IsMe(acptr))		/* no es para mi, a rutar */
      {
	sendto_one(acptr, ":%s DBQ %s %c %s", parv[0], servidor, tabla, clave);
	return 0;		/* ok, rutado, fin del trabajo */
      }
    }
  }

  if (!tabla_residente[tabla])
  {
    if (MyUser(sptr) || Protocol(cptr) < 10)
      sendto_one(cptr,
	  ":%s NOTICE %s :DBQ ERROR Tabla='%c' Clave='%s' TABLA_NO_RESIDENTE",
	  me.name, parv[0], tabla, clave);
    else
      sendto_one(cptr,
	  "%s NOTICE %s%s :DBQ ERROR Tabla='%c' Clave='%s' TABLA_NO_RESIDENTE",
	  NumServ(&me), NumNick(sptr), tabla, clave);
    return 0;
  }

/*
** Limitamos el acceso a ciertas tablas y ciertos registros
*/
  if (!IsServer(sptr) && IsHelpOp(sptr))
  {
    reg = db_buscar_registro(BDD_OPERDB, sptr->name);
    if (reg)
    {
      nivel_helper = atoi(reg->valor);
    }
  }

  switch (tabla)
  {
    case ESNET_NICKDB:		/* 'n' */
      if (nivel_helper < 10)
	nivel_helper = -1;
      break;
    case BDD_IPVIRTUALDB:	/* 'v' */
      if (!strcmp(clave, ".") && (nivel_helper < 10))
	nivel_helper = -1;
      break;
  }

  if (nivel_helper < 0)
  {
    if (MyUser(sptr) || Protocol(cptr) < 10)
      sendto_one(cptr,
	  ":%s NOTICE %s :DBQ ERROR No tienes permiso para acceder a Tabla='%c' Clave='%s'",
	  me.name, parv[0], tabla, clave);
    else
      sendto_one(cptr,
	  "%s NOTICE %s%s :DBQ ERROR No tienes permiso para acceder a Tabla='%c' Clave='%s'",
	  NumServ(&me), NumNick(sptr), tabla, clave);
    return 0;
  }

  reg = db_buscar_registro(tabla, clave);
  if (!reg)
  {
    if (MyUser(sptr) || Protocol(cptr) < 10)
      sendto_one(cptr,
	  ":%s NOTICE %s :DBQ ERROR Tabla='%c' Clave='%s' REGISTRO_NO_ENCONTRADO",
	  me.name, parv[0], tabla, clave);
    else
      sendto_one(cptr,
	  "%s NOTICE %s%s :DBQ ERROR Tabla='%c' Clave='%s' REGISTRO_NO_ENCONTRADO",
	  NumServ(&me), NumNick(sptr), tabla, clave);
    return 0;
  }


  if (MyUser(sptr) || Protocol(cptr) < 10)
    sendto_one(cptr,
	":%s NOTICE %s :DBQ OK Tabla='%c' Clave='%s' Valor='%s'",
	me.name, parv[0], tabla, reg->clave, reg->valor);
  else
    sendto_one(cptr,
	"%s NOTICE %s%s :DBQ OK Tabla='%c' Clave='%s' Valor='%s'",
	NumServ(&me), NumNick(sptr), tabla, reg->clave, reg->valor);
  return 0;
}

#endif /* BDD */

#endif /* S_BDD.C */
